STATUS_PHRASES = {
    200: "OK",
    201: "Created",
    202: "Accepted",
    204: "No Content",
    301: "Moved Permanently",
    302: "Found",
    400: "Bad Request",
    401: "Unauthorized",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    409: "Conflict",
    422: "Unprocessable Entity",
    500: "Internal Server Error",
    502: "Bad Gateway",
    503: "Service Unavailable",
}



class Config:
    def __init__(self, user_config=None):
        self._defaults = {
            "UPLOAD_DIR": "uploads",         # where to save uploaded files
            "DEBUG": False,                  # show internal stacktrace
            "LOGGING_ENABLED": False,        # enable logging (later)
        }
        self._overrides = user_config or {}

    def get(self, key, default=None):
        return self._overrides.get(key, self._defaults.get(key, default))

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        self._overrides[key] = value

    def __contains__(self, key):
        return key in self._overrides or key in self._defaults

    def to_dict(self):
        data = self._defaults.copy()
        data.update(self._overrides)
        return data



import inspect
from functools import wraps
from slush.core.routing import Router
from slush.core.request import Request
from slush.core.response import Response
from slush.core.exception import HTTPBaseException
from slush.exception import NotFound, InternalServerError


class Slush:
    def __init__(self):
        self.router = Router()

    def route(self, path, methods=['GET']):
        """
        Decorator to register a route with the application.
        :param path: The URL path for the route.
        :param methods: List of HTTP methods this route responds to.
        :return: Decorator function.
        """
        if isinstance(methods, str):
            methods = [methods.upper()]
        if not all(method in ['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'OPTIONS'] for method in methods):
            raise ValueError("Invalid HTTP method specified. Allowed methods are: GET, POST, PUT, DELETE, PATCH, OPTIONS.")
        if not path.startswith('/'):
            raise ValueError("Path must start with a '/' character.")
        if not path:
            raise ValueError("Path cannot be empty.")

        def decorator(func):
            @wraps(func)
            def wrapped_func(*args, **kwargs):
                return func(*args, **kwargs)

            for method in methods:
                self.router.add_route(path, method, wrapped_func)

            return wrapped_func

        return decorator

    def __call__(self, environ, start_response):
        """
        WSGI application entry point.
        :param environ: WSGI environment dictionary.
        :param start_response: WSGI start response callable.
        :return: WSGI response iterable.
        """

        request = Request(environ)
        handler, path_params = self.router.resolve(request.path, request.method)

        try:
            if not handler:
                raise NotFound()

            # Inject request and path params
            sig = inspect.signature(handler)
            kwargs = {}

            if "request" in sig.parameters:
                kwargs["request"] = request
            for name in path_params:
                if name in sig.parameters:
                    kwargs[name] = path_params[name]

            result = handler(**kwargs)
            if isinstance(result, Response):
                response = result
            else:
                response = Response(body=result)

            status, headers, body = response.finalize()
            start_response(status, headers)
            return body

        except HTTPBaseException as http_exc:
            response = http_exc.as_response()
            start_response(response.status, response.headers)
            return [response.body]

        except Exception as e:
            error = InternalServerError(str(e))
            response = error.as_response()
            start_response(response.status, response.headers)
            return [response.body]




import re
from typing import Callable, Dict, Tuple, Any
from slush.exception import NotFound, MethodNotAllowed

class Route:
    def __init__(self, path: str, method: str, handler: Callable):
        """
        Initialize a Route object.
        :param path: The URL path for the route.
        :param method: The HTTP method (GET, POST, etc.) for the route.
        :param handler: The function that handles requests to this route.
        """
        self.path = path
        self.method = method.upper()
        self.handler = handler
        self.path_regex, self.param_names = self._compile_path(path)

    def _compile_path(self, path: str):
        # Convert /user/<int:id>/ to regex pattern
        """
        Compile the path into a regex pattern and extract parameter names.
        :param path: The URL path to compile.
        :return: A tuple containing the compiled regex pattern and a list of parameter names.
        """
        pattern = "^"
        param_names = []

        tokens = path.strip("/").split("/")
        for token in tokens:
            if token.startswith("<") and token.endswith(">"):
                type_and_name = token[1:-1].split(":")
                if len(type_and_name) != 2:
                    raise ValueError("Invalid dynamic path: " + token)
                type_, name = type_and_name
                param_names.append((name, type_))
                if type_ == "int":
                    pattern += r"/(?P<%s>\d+)" % name
                elif type_ == "str":
                    pattern += r"/(?P<%s>[^/]+)" % name
                else:
                    raise ValueError(f"Unsupported path type: {type_}")
            else:
                pattern += f"/{token}"

        pattern += "/?$"
        return re.compile(pattern), param_names

    def match(self, path: str):
        """
        Match the given path against the route's regex pattern.
        :param path: The URL path to match.
        :return: A dictionary of extracted parameters if the path matches, otherwise None.
        """
        match = self.path_regex.match(path)
        if not match:
            return None
        return match.groupdict()

class Router:
    """A singleton router that manages routes and resolves them based on the request path and method.
    """
    def __init__(self):
        self.routes: list[Route] = []

    def add_route(self, path: str, method: str, handler: Callable):
        self.routes.append(Route(path, method, handler))

    def resolve(self, path: str, method: str) -> Tuple[Callable, Dict[str, Any]]:
        allowed_methods = []

        for route in self.routes:
            params = route.match(path)
            if params is not None:
                allowed_methods.append(route.method)
                if route.method == method:
                    return route.handler, params

        if allowed_methods:
            raise MethodNotAllowed(
                detail=f"Allowed methods: {', '.join(allowed_methods)}",
                allowed_methods=allowed_methods
            )

        raise NotFound()



# core/request.py

import json
from urllib.parse import parse_qs
from cgi import FieldStorage
from typing import Optional
from slush.utils.files import UploadedFile


class Request:
    def __init__(self, environ: dict):
        """
        Initialize a Request object from WSGI environment.
        """
        self.environ = environ
        self.method = environ["REQUEST_METHOD"]
        self.path = environ["PATH_INFO"]
        self._body = None  # Lazy-loaded

        self.query_params = self._parse_query()
        self.headers = self._parse_headers()
        self.cookies = self._parse_cookies()

    def _parse_query(self) -> dict:
        raw_query = self.environ.get("QUERY_STRING", "")
        return {
            k: v[0] if len(v) == 1 else v
            for k, v in parse_qs(raw_query).items()
        }

    def _parse_headers(self) -> dict:
        headers = {}
        for key, value in self.environ.items():
            if key.startswith("HTTP_"):
                header = key[5:].replace("_", "-").title()
                headers[header] = value
        if "CONTENT_TYPE" in self.environ:
            headers["Content-Type"] = self.environ["CONTENT_TYPE"]
        if "CONTENT_LENGTH" in self.environ:
            headers["Content-Length"] = self.environ["CONTENT_LENGTH"]
        return headers

    def _parse_cookies(self) -> dict:
        cookie_str = self.environ.get("HTTP_COOKIE", "")
        cookies = {}
        for pair in cookie_str.split(";"):
            if "=" in pair:
                k, v = pair.strip().split("=", 1)
                cookies[k] = v
        return cookies

    def _read_body(self) -> str:
        if self._body is None:
            try:
                length = int(self.environ.get("CONTENT_LENGTH", 0))
            except (ValueError, TypeError):
                length = 0
            self._body = (
                self.environ["wsgi.input"].read(length).decode("utf-8")
                if length > 0
                else ""
            )
        return self._body

    @property
    def body(self) -> str:
        return self._read_body()

    def json(self) -> Optional[dict]:
        try:
            return json.loads(self._read_body())
        except (json.JSONDecodeError, UnicodeDecodeError):
            return None

    def form(self) -> dict:
        self._parse_form_data()
        return {
            key: fs.getvalue()
            for key, fs in self._form_data.items()
            if not isinstance(fs, FieldStorage) or not fs.filename
        }

    def files(self) -> dict:
        self._parse_form_data()
        return {
            key: UploadedFile(fs)
            for key, fs in self._form_data.items()
            if isinstance(fs, FieldStorage) and fs.filename
        }

    def _parse_form_data(self):
        if hasattr(self, "_form_data"):
            return  # already parsed

        self._form_data = {}
        try:
            fs = FieldStorage(
                fp=self.environ["wsgi.input"],
                environ=self.environ,
                keep_blank_values=True
            )
            for key in fs:
                self._form_data[key] = fs[key]
        except Exception:
            self._form_data = {}


from slush.core.exception import HTTPBaseException

class NotFound(HTTPBaseException):
    def __init__(self, detail="Not Found"):
        super().__init__(404, detail)


class MethodNotAllowed(HTTPBaseException):
    def __init__(self, detail="Method Not Allowed", allowed_methods=None):
        headers = [("Content-Type", "application/json")]
        if allowed_methods:
            headers.append(("Allow", ", ".join(allowed_methods)))
        super().__init__(405, detail, headers)


class BadRequest(HTTPBaseException):
    def __init__(self, detail="Bad Request"):
        super().__init__(400, detail)


class InternalServerError(HTTPBaseException):
    def __init__(self, detail="Internal Server Error"):
        super().__init__(500, detail)


class ValidationError(HTTPBaseException):
    def __init__(self, detail="Validation Error"):
        super().__init__(422, detail)


# core/exceptions.py

class HTTPBaseException(Exception):
    def __init__(self, status_code: int, detail: str = "", headers=None):
        self.status_code = status_code
        self.detail = detail or self.default_detail()
        self.headers = headers or [("Content-Type", "application/json")]

    def default_detail(self):
        return "An error occurred"

    def as_response(self):
        from slush.core.response import Response
        return Response(
            body={"error": self.detail},
            status=f"{self.status_code} {self.status_text()}",
            headers=self.headers
        )

    def status_text(self):
        return {
            400: "Bad Request",
            401: "Unauthorized",
            403: "Forbidden",
            404: "Not Found",
            405: "Method Not Allowed",
            422: "Unprocessable Entity",
            500: "Internal Server Error",
        }.get(self.status_code, "Unknown Error")



# slush/server.py
import os
import sys
from wsgiref.simple_server import make_server
from threading import Thread

def run(app, host="127.0.0.1", port=8000, debug=False):
    """
    Run a WSGI application with optional auto-reloading using Watchdog.
    :param app: WSGI application instance
    :param host: Hostname to bind to
    :param port: Port number to bind to
    :param debug: Enables auto-reloading on file changes
    """
    if debug:
        try:
            from watchdog.observers import Observer
            from watchdog.events import FileSystemEventHandler
        except ImportError:
            raise ImportError("⚠️ Auto-reload requires 'watchdog'. Install it using 'pip install watchdog'.")

        class ReloadHandler(FileSystemEventHandler):
            def on_modified(self, event):
                if event.src_path.endswith(".py"):
                    print(f"🔄 Detected change in {event.src_path}. Restarting...")
                    observer.stop()
                    os.execv(sys.executable, [sys.executable] + sys.argv)

        observer = Observer()
        observer.schedule(ReloadHandler(), path='.', recursive=True)
        observer.start()

        print(f"🚀 Slush server started with auto-reload: http://{host}:{port}")
        try:
            _start_server(app, host, port)
        finally:
            observer.stop()
            observer.join()
    else:
        print(f"🚀 Slush server running on http://{host}:{port}")
        _start_server(app, host, port)


def _start_server(app, host, port):
    with make_server(host, port, app) as server:
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            print("\n❌ Server stopped manually.")



import json
from slush.const import STATUS_PHRASES

class Response:
    def __init__(self, body="", status=200, headers=None, content_type=None):
        self.headers = headers or []
        self._cookies = []

        if isinstance(status, str) and status.isdigit():
            status = int(status)

        if isinstance(status, int):
            phrase = STATUS_PHRASES.get(status, "OK")
            self.status = f"{status} {phrase}"
        else:
            self.status = status

        # Content type inference
        if isinstance(body, (dict, list)):
            body = json.dumps(body)
            content_type = content_type or "application/json"
        elif isinstance(body, str):
            content_type = content_type or "text/plain"
        else:
            content_type = content_type or "application/octet-stream"

        if isinstance(body, str):
            body = body.encode("utf-8")

        self.body = body
        if content_type is not None:
            self.set_header("Content-Type", content_type)   

    def set_header(self, key, value):
        self.headers = [h for h in self.headers if h[0].lower() != key.lower()]
        self.headers.append((key, value))

    def set_cookie(self, key, value, path="/", httponly=True, max_age=None):
        cookie = f"{key}={value}; Path={path}"
        if httponly:
            cookie += "; HttpOnly"
        if max_age is not None:
            cookie += f"; Max-Age={max_age}"
        self._cookies.append(("Set-Cookie", cookie))

    def finalize(self):
        if not isinstance(self.body, bytes):
            self.body = str(self.body).encode("utf-8")
        return self.status, self.headers + self._cookies, [self.body]

    @classmethod
    def json(cls, data, status=200, headers=None):
        body = json.dumps(data)
        return cls(body=body, status=status, headers=headers, content_type="application/json")

    @classmethod
    def text(cls, data, status=200, headers=None):
        return cls(body=data, status=status, headers=headers, content_type="text/plain")

    @classmethod
    def html(cls, data, status=200, headers=None):
        return cls(body=data, status=status, headers=headers, content_type="text/html")

    @classmethod
    def redirect(cls, location, status=302): 
        return cls(body=b"", status=status, headers=[("Location", location)], content_type=None)



from cgi import FieldStorage, MiniFieldStorage
from typing import Optional, BinaryIO

class UploadedFile:
    def __init__(self, field: FieldStorage):
        self.filename: str = field.filename
        self.content_type: str = field.type
        self.file: BinaryIO = field.file
        self.size: int = self._get_file_size()

    def _get_file_size(self) -> int:
        current_pos = self.file.tell()
        self.file.seek(0, 2)  # Seek to end
        size = self.file.tell()
        self.file.seek(current_pos)  # Return to original position
        return size

    def read(self) -> bytes:
        self.file.seek(0)
        return self.file.read()

    def save(self, path: str):
        with open(path, "wb") as f:
            f.write(self.read())